#SequenceAnalysis.pm - Collection of basic functions to do routine sequence manipulation 
#1)Calculate bp composition of given sequence
#2)Calculate the composition % for each bases at position +1, +2, +3
#3)Reverse complements a DNA sequence
#4)Translate dna sequence into amino acid sequence using the universal aa code
#Rene Warren
#October 2002
package SequenceAnalysis;

##############################
# perldoc_header             #
##############################

=head1 NAME <UPLINK>

SequenceAnalysis.pm - SequenceAnalysis.pm - Collection of basic functions to do routine sequence manipulation 

=head1 SYNOPSIS <UPLINK>

	<<SYNOPSIS>>

=head1 DESCRIPTION <UPLINK>

=for html
SequenceAnalysis.pm - Collection of basic functions to do routine sequence manipulation <BR>1)Calculate bp composition of given sequence<BR>2)Calculate the composition % for each bases at position +1, +2, +3<BR>3)Reverse complements a DNA sequence<BR>4)Translate dna sequence into amino acid sequence using the universal aa code<BR>Rene Warren<BR>October 2002<BR>

=cut

##############################
# superclasses               #
##############################

=head2 SUPERCLASSES <UPLINK>

=over

=back

=cut

##############################
# system_variables           #
##############################
##############################
# standard_modules_ref       #
##############################

=head2 STANDARD MODULES REFERENCE <UPLINK>

=over

=item strict

=item Number::Format

=back

=cut

use strict;
use Number::Format;

##############################
# custom_modules_ref         #
##############################

=head2 CUSTOM MODULES REFERENCE <UPLINK>

=over

=back

=cut

##############################
# global_vars                #
##############################

=head2 GLOBAL VARIABLES <UPLINK>

=over

=back

=cut

##############################
# modular_vars               #
##############################
##############################
# constants                  #
##############################
##############################
# main_header                #
##############################
my $format= new Number::Format(-thousands_sep => ',');1;

##############################
# constructor                #
##############################
##############################
# public_methods             #
##############################
##############################
# public_functions           #
##############################

=head2 PUBLIC FUNCTIONS <UPLINK>

=for html <font size='-1'><strong><A href='#countBase'>countBase; </A></strong></font><BR>

=for html <font size='-1'><strong><A href='#calculateComposition'>calculateComposition; </A></strong></font><BR>

=for html <font size='-1'><strong><A href='#compositionPosition'>compositionPosition; </A></strong></font><BR>

=for html <font size='-1'><strong><A href='#reverseComplement'>reverseComplement; </A></strong></font><BR>

=for html <font size='-1'><strong><A href='#dna2protein'>dna2protein; </A></strong></font><BR>

=for html <font size='-1'><strong><A href='#sharedAA'>sharedAA; </A></strong></font><BR>

=over

=for html <A NAME='countBase'>

=item countBase

=for html
simply counts number of each bases in a sequence and consensus if sequence<BR>bases are in caps<BR>

=for html
<B>Arguments:</B><BR>
********************<BR>
<Font size =-2>- <B>$dna_sequence</B>():  [Default = <B>shift</B>]<BR>
</Font><BR>

=cut

#---------------------------------------------------
sub countBase{

   #simply counts number of each bases in a sequence and consensus if sequence
   #bases are in caps
   my $dna_sequence=shift;

   my @base;
   my $consensus;
   my ($counta, $countt, $countg, $countc);

   while ($dna_sequence =~ /(a|c|g|t|x|n)/gi){
        $consensus++ if ($1 eq 'A' or $1 eq 'T' or $1 eq 'G' or $1 eq 'C');

        my $single_base=lc($1);

        $counta++ if ($single_base eq 'a');
        $countt++ if ($single_base eq 't');
        $countg++ if ($single_base eq 'g');
        $countc++ if ($single_base eq 'c');
   }

  @base=($counta,$countt,$countg,$countc);

  return $consensus, \@base;
}

=for html <A NAME='calculateComposition'>

=item calculateComposition

=for html
Calculate bp composition of given sequence<BR>

=for html
<B>Arguments:</B><BR>
********************<BR>
<Font size =-2>- <B>$seq</B>():  [Default = <B>shift</B>]<BR>
</Font><BR>

=cut

#---------------------------------------------------
sub calculateComposition{

    #Calculate bp composition of given sequence

    my $seq=shift;

    my ($counta, $countt, $countc, $countg);
    my ($consensus, $base);
    my ($perca, $perct, $percc, $percg, $percgc);
    my $l;

    $l+=length($seq);

    ($consensus, $base)=&countBase($seq);
    my @base=@$base;    

    my $total_bases=$base[0]+$base[1]+$base[2]+$base[3];

    if ($total_bases ne '0'){
        $perca=$base[0]/$total_bases*100;
        $perct=$base[1]/$total_bases*100;
        $percg=$base[2]/$total_bases*100;
        $percc=$base[3]/$total_bases*100;
    }

    $percgc=$percg+$percc;

    return ($perca, $perct, $percg, $percc, $percgc);
}

=for html <A NAME='compositionPosition'>

=item compositionPosition

=for html
given a sequence from a string,<BR>Calculate the composition % for each bases at position +1, +2, +3<BR>

=for html
<B>Arguments:</B><BR>
********************<BR>
<Font size =-2>- <B>$seq</B>():  [Default = <B>shift</B>]<BR>
</Font><BR>

=cut

#---------------------------------------------------
sub compositionPosition{
   #given a sequence from a string,
   #Calculate the composition % for each bases at position +1, +2, +3

    my $seq=shift;

    $seq=uc($seq);
    my @temp_array=split(//,$seq);

    my $total;
    my $element;
    my %pos_composition;

    for (my $x=1;$x<=3;$x++){
        $element=$x-1;
        my $a=$element."a";
        my $t=$element."t";
        my $g=$element."g";
        my $c=$element."c";

        for (my $i=$element;$i<$#temp_array; $i+=3){

            $a++ if ($temp_array[$i] eq 'A');
            $t++ if ($temp_array[$i] eq 'T');
            $g++ if ($temp_array[$i] eq 'G');
            $c++ if ($temp_array[$i] eq 'C');
        }
        $total=$a+$t+$g+$c;
        $a=$format->format_number($a/$total*100) if ($total ne '0');
        $t=$format->format_number($t/$total*100) if ($total ne '0');
        $g=$format->format_number($g/$total*100) if ($total ne '0');
        $c=$format->format_number($c/$total*100) if ($total ne '0');

        #print "$a\t$t\t$g\t$c\n";

        $pos_composition{$x}{'a'}=$a;
        $pos_composition{$x}{'t'}=$t;
        $pos_composition{$x}{'g'}=$g;
        $pos_composition{$x}{'c'}=$c;

        $total=0; ###resets the total at that position
    }
    #print "@compositionarray\n";

    return \%pos_composition;
}

=for html <A NAME='reverseComplement'>

=item reverseComplement

=for html
Reverse complements a DNA sequence<BR>only accepts sequence as string<BR>returns the reverse complemented sequence in a string<BR>

=for html
<B>Arguments:</B><BR>
********************<BR>
<Font size =-2>- <B>$seq</B>():  [Default = <B>shift</B>]<BR>
</Font><BR>

=cut

#---------------------------------------------------
sub reverseComplement{

    #Reverse complements a DNA sequence
    #only accepts sequence as string
    #returns the reverse complemented sequence in a string

    my $seq=shift;
    $seq=lc($seq);

    my @temp_array={};
    @temp_array=split(//,$seq);
    my @temp_array1=reverse(@temp_array);

    my $rev_seq="";

    #reverse complements
    my %comprev;
    $comprev{"g"}= "c";  $comprev{"c"}= "g"; #Define Reverse Matrix
    $comprev{"a"}= "t";  $comprev{"t"}= "a";

    for (my $i=0; $i<=$#temp_array1; $i+=1){
        my $comp=$temp_array1[$i];
        if (defined $comprev{$comp}){
            $rev_seq.=$comprev{$comp};
        }
        #undecidable nucleic acid
        else{
            $rev_seq.="n";
        }

    }

    return $rev_seq;

}

=for html <A NAME='dna2protein'>

=item dna2protein

=for html
converts the plain nucleotide sequence to amino acid sequence<BR>

=for html
<B>Arguments:</B><BR>
********************<BR>
<Font size =-2>- <B>$seq</B>():  [Default = <B>shift</B>]<BR>
</Font><BR>

=cut

#---------------------------------------------------
sub dna2protein{
# converts the plain nucleotide sequence to amino acid sequence

    my $seq=shift;
    $seq=lc($seq);
    my @temp_array=split(//,$seq);

    my %codon2aa;

    $codon2aa{"ttt"}= "F"; $codon2aa{"ttc"}= "F"; $codon2aa{"tta"}="L";
    $codon2aa{"ttg"}= "L"; $codon2aa{"tct"}= "S"; $codon2aa{"tcc"}= "S";
    $codon2aa{"tca"}= "S"; $codon2aa{"tcg"}= "S"; $codon2aa{"tat"}= "Y";
    $codon2aa{"tac"}= "Y"; $codon2aa{"tgt"}= "C"; $codon2aa{"tgc"}= "C";
    $codon2aa{"ctt"}= "L"; $codon2aa{"ctc"}= "L"; $codon2aa{"cta"}= "L";
    $codon2aa{"ctg"}= "L"; $codon2aa{"cct"}= "P"; $codon2aa{"ccc"}= "P";
    $codon2aa{"cca"}= "P"; $codon2aa{"ccg"}= "P"; $codon2aa{"cat"}= "H";
    $codon2aa{"cac"}= "H"; $codon2aa{"caa"}= "Q"; $codon2aa{"cag"}= "Q";
    $codon2aa{"cgt"}= "R"; $codon2aa{"cgc"}= "R"; $codon2aa{"cga"}= "R";
    $codon2aa{"cgg"}= "R"; $codon2aa{"att"}= "I"; $codon2aa{"atc"}= "I";
    $codon2aa{"ata"}= "I"; $codon2aa{"atg"}= "M"; $codon2aa{"act"}= "T";
    $codon2aa{"acc"}= "T"; $codon2aa{"aca"}= "T"; $codon2aa{"acg"}= "T";
    $codon2aa{"aat"}= "N"; $codon2aa{"aac"}= "N"; $codon2aa{"aaa"}= "K";
    $codon2aa{"aag"}= "K"; $codon2aa{"agt"}= "S"; $codon2aa{"agc"}= "S";
    $codon2aa{"aga"}= "R"; $codon2aa{"agg"}= "R"; $codon2aa{"gtt"}= "V";
    $codon2aa{"gtc"}= "V"; $codon2aa{"gta"}= "V"; $codon2aa{"gtg"}= "V";
    $codon2aa{"gct"}= "A"; $codon2aa{"gcc"}= "A"; $codon2aa{"gca"}= "A";
    $codon2aa{"gcg"}= "A"; $codon2aa{"gat"}= "D"; $codon2aa{"gac"}= "D";
    $codon2aa{"gaa"}= "E"; $codon2aa{"gag"}= "E"; $codon2aa{"ggt"}= "G";
    $codon2aa{"ggc"}= "G"; $codon2aa{"gga"}= "G"; $codon2aa{"ggg"}= "G";
    $codon2aa{"tgg"}= "W";
    $codon2aa{"tag"}= "*";
    $codon2aa{"tga"}= "*";
    $codon2aa{"taa"}= "*";

    my $aa_seq="";

    for (my $pos=0; $pos<$#temp_array; $pos+=3){
        my $codon=$temp_array[$pos].$temp_array[$pos+1].$temp_array[$pos+2];
        if (defined $codon2aa{$codon}){
            $aa_seq.=$codon2aa{$codon};
        }
        #undecidable nucleic acid
        else{
            $aa_seq.="X";
        }
    }
    return $aa_seq;
}

=for html <A NAME='sharedAA'>

=item sharedAA

=for html
generate a sequence containing aa similarity between 2 protein sequences<BR>

=for html
<B>Arguments:</B><BR>
********************<BR>
- None<BR>
<BR>

=cut

#-------------------------------------
sub sharedAA{

    #generate a sequence containing aa similarity between 2 protein sequences
    my ($query_sequence, $hit_sequence)=@_;

    my @query=split(//,$query_sequence);       #put sequences in arrays
    my @hit=split(//, $hit_sequence);

    my $shared;
    #based on similar Chemical properties between amino acids

    for (my $p;$p<=$#query;$p++){     #p is position
        if ($query[$p] eq $hit[$p]){  #same aa
            $shared .= $query[$p];
        }
        #VILM
        elsif (($query[$p] eq 'V' && ($hit[$p] eq 'I' or $hit[$p] eq 'L' or $hit[$p] eq 'M')) or ($hit[$p] eq 'V' && ($query[$p] eq 'I' or $query[$p] eq 'L' or $query[$p] eq 'M'))){
            $shared .= '+';
        }
        elsif (($query[$p] eq 'I' &&  ($hit[$p] eq 'L' or $hit[$p] eq 'M')) or ($hit[$p] eq 'I' && ($query[$p] eq 'L' or $query[$p] eq 'M'))){
            $shared .= '+';
        }
        elsif (($query[$p] eq 'L' &&  $hit[$p] eq 'M') or ($hit[$p] eq 'L' && $query[$p] eq 'M')){
            $shared .= '+';
        }

        #DNE
        elsif (($query[$p] eq 'D' && ($hit[$p] eq 'N' or $hit[$p] eq 'E')) or ($hit[$p] eq 'D' && ($query[$p] eq 'N' or $query[$p] eq 'E'))){
            $shared .= '+';
        }
        elsif (($query[$p] eq 'E' && ($hit[$p] eq 'Q' or $hit[$p] eq 'D' or $hit[$p] eq 'K')) or ($hit[$p] eq 'E' && ($query[$p] eq 'Q' or $query[$p] eq 'D' or $query[$p] eq 'K'))){
            $shared .= '+';
        }
        elsif (($query[$p] eq 'S' && $hit[$p] eq 'T' ) or ($hit[$p] eq 'S' && $query[$p] eq 'T')){
            $shared .= '+';
        }
        elsif (($query[$p] eq 'R' && ($hit[$p] eq 'K' or $hit[$p] eq 'Q')) or ($hit[$p] eq 'R' && ($query[$p] eq 'K' or $query[$p] eq 'Q'))){
            $shared .= '+';
        }
        elsif (($query[$p] eq 'K' && $hit[$p] eq 'Q') or ($hit[$p] eq 'K' && $query[$p] eq 'Q')){
            $shared .= '+';
        }
        elsif (($query[$p] eq 'S' && $hit[$p] eq 'A') or ($hit[$p] eq 'S' && $query[$p] eq 'A')){
            $shared.= '+';
        }
        elsif (($query[$p] eq 'S' && $hit[$p] eq 'N') or ($hit[$p] eq 'S' && $query[$p] eq 'N')){
            $shared.= '+';
        }
  elsif (($query[$p] eq 'Y' && $hit[$p] eq 'H') or ($hit[$p] eq 'Y' && $query[$p] eq 'H')){
            $shared.= '+';
        }
        elsif (($query[$p] eq 'H' && $hit[$p] eq 'N') or ($hit[$p] eq 'H' && $query[$p] eq 'N')){
            $shared.= '+';
        }
        #NO MATCH/NO HOMOLOGY
        else{
            $shared.= ' ';
        }
    }
    return $shared;
}

=back

=cut

##############################
# private_methods            #
##############################
##############################
# private_functions          #
##############################
##############################
# main_footer                #
##############################
##############################
# perldoc_footer             #
##############################

=head1 KNOWN ISSUES <UPLINK>

<<KNOWN ISSUES>>

=head1 FUTURE IMPROVEMENTS <UPLINK>

<<FUTURE IMPROVEMENTS>>

=head1 AUTHORS <UPLINK>

<<AUTHORS>>

=head1 CREATED <UPLINK>

2004-08-18

=head1 REVISION <UPLINK>


=cut


return 1;
