#Rene Warren
#January 2003, June 2003
#XMLparser.pm module

=head1 NAME
   XMLparser - A module for specifically populating SAMdb with data coming from xml files (assembly.xml & summary.xml)

=head1 SYNOPSIS
   use XMLparser;
   XMLparser::assemblyData($user_info, $xml_dir, $assembly_id, $configuration{'Config_AssemblyXml_File'}, $date_condition);
   XMLparser::summaryData($user_info, $xml_dir, $assembly_id, $configuration{'Config_SummaryXml_File'}, $read_analyzed);

   -the first argument is a reference to a hash containing the user info in order to connect to the database:

   $user_info{'username'}=$opt_u;
   $user_info{'password'}=$opt_p;
   $user_info{'host_name'}=$opt_o;
   $user_info{'database'}=$opt_d;

   -the second argument is a string containing the fullpath where the xml files are strored

   -the 3rd argument is the current assembly id (not build id)
   
   -the 4th argument is the name of the file to parse into the database

   -the 5th argument is either the date condition
   e.g.
   Trace_Date_Sequenced>=\'$date_start\' AND Trace_Date_Sequenced<=\'$date_end 23:59:59\'

   or the number of reads analyzed (data type is number like the 3rd argument)
   

=head1 DESCRIPTION
   This module, along with XML::Simple will parse an assembly and summary xml document and parse its content into any SAMdb

=head1 EXAMPLES

=head1 NOTES

=head1 AUTHOR
   Rene Warren April 2003, June 2003
=cut



package XMLparser;

use strict;
use Date::Manip;
use XML::Simple;
use SQLdb; 
use Data::Dumper;


###################################################################### fingerprint_map.xml parser ######################################################
#-------------------------------------
sub readMapData{

   my $user_info=shift;
   my %user_info=%$user_info;
   my $xml_file=shift;
   my $fosmidboundary_size=shift; 
  
   my ($status, $message);
   my $dbh;

   eval{

      ####Initialize Timer
      my $start_date=&getDate();
   
      ####Get XML data into hashref
      print "Parsing Assembly XML...\n";
      my $xml_record=XMLin($xml_file, forcearray=>['read_name'], suppressempty=>1);#read_name will always be in array

#      print Dumper($xml_record);
#      &testStructure($xml_record);

      ####Initialize database connection
      print "Connecting to $user_info{'database'} database...\n";
      $dbh=SQLdb::Initialize($user_info{'password'}, $user_info{'username'}, $user_info{'database'}, $user_info{'host_name'});

      ####Get Map, Map_Contig and Map_Layout initial values
      print "Getting Initial db values...\n";
      my ($map_record, $maptig_record, $clone_info, $enzyme_info)=&getInitialMapValue($dbh, $fosmidboundary_size);

      ####Parse XML and Fill SAMdb
      print "Populating $user_info{'database'} database...\n";
      my ($map_insert_ct, $maptig_insert_ct, $layout_insert_ct)=&parseMapXMLStructure($map_record, $maptig_record, $clone_info, $enzyme_info, $xml_record, $user_info{'database'}, $dbh);

      ####End Time
      my $end_date=&getDate();

      ####Time Elapsed
      my $time_elapsed=&DateCalc($start_date,$end_date);

      print "*started at    :$start_date\n";
      print "*Finished at   :$end_date\n";
      print "*Time Elapsed  :$time_elapsed\n";

      print "Map Inserted:$map_insert_ct\nMap Contig Inserted:$maptig_insert_ct\nMap Layout Inserted:$layout_insert_ct\n"

   };
   if($@){$status="failed";$message="$@";print "$@\n";SQLdb::Disconnect($dbh);}else{$status="completed";print "Success!\n";SQLdb::Disconnect($dbh);}


   return $status, $message;
}

#-------------------------------------
#Get all current values
#not so necessary since we use the auto-increment function of mysql
sub getInitialMapValue{

   my $dbh=shift;
   my $fosmidboundary_size=shift;

   my @map_field=("count(Map_ID)");
   my @map_table=("Map");
   my @map_condition=();
   my $select_map=SQLdb::GetRecord(\@map_field, \@map_table, \@map_condition);
   my $map_record=SQLdb::ExecuteUnique($dbh, $select_map);

   my @maptig_field=("count(Map_Contig_ID)");
   my @maptig_table=("Map_Contig");
   my @maptig_condition=();
   my $select_maptig=SQLdb::GetRecord(\@maptig_field, \@maptig_table, \@maptig_condition);
   my $maptig_record=SQLdb::ExecuteUnique($dbh, $select_maptig);

   ######GET CLONE_ID AND NAME RELATIONSHIP (can't make use of clone date time sequenced, neither specific library name...will try based on Library Size)
   ####To cover a whole genome with a minimal number of tiling clones, you need fairly large clones from BAC or fosmid libraries
   ####Hopefully, 10k insert size is ideal to segregate plasmid vs fosmid clones ---if not, change value here
   #select Clone_ID, Clone_Name from Clone, Library where Library_ID=FK_Library__ID and Library_Insert_Size>10000;

   my @clone_field=("Clone_Name", "Clone_ID");
   my @clone_table=("Clone", "Library");
   my @clone_condition=("Library_ID=FK_Library__ID", "Library_Insert_Size>$fosmidboundary_size");###fosmidboundary_size value above in this program
   my $select_clone=SQLdb::GetRecord(\@clone_field, \@clone_table, \@clone_condition);
   my $multiple_clone=SQLdb::ExecuteMany($dbh, $select_clone);
   my @clone_record=@$multiple_clone;
   my %clone_info;

   #if there's no clone it is not a big deal.  It only means that map clones will NOT be linked to the Clone table for that fingerprint map version

   foreach my $clone_rec(@clone_record){
      my @clone_data=@$clone_rec;
      $clone_info{$clone_data[0]}=$clone_data[1];
   }

   ######GET ENZYME_ID AND NAME RELATIONSHIP
   my @enzyme_field=("Enzyme_Name", "Enzyme_ID");
   my @enzyme_table=("Enzyme");
   my @enzyme_condition=();
   my $select_enzyme=SQLdb::GetRecord(\@enzyme_field, \@enzyme_table, \@enzyme_condition);
   my $multiple_enzyme=SQLdb::ExecuteMany($dbh, $select_enzyme);
   my @enzyme_record=@$multiple_enzyme;
   my %enzyme_info;

   foreach my $enzyme_rec(@enzyme_record){
      my @enzyme_data=@$enzyme_rec;
      my $uppercase_enzymename=uc($enzyme_data[0]);       ###ensures that a link will be created in the database regardless of naming convention not being respected
      print "$uppercase_enzymename .... $enzyme_data[1]\n";
      $enzyme_info{$uppercase_enzymename}=$enzyme_data[1];
   }


   return $map_record, $maptig_record, \%clone_info, \%enzyme_info;
}

#-------------------------------------
sub checkMapExistence{

    my $map_version=shift;
    my $dbh=shift;

    #CHECK EXISTENCE OF MAP
    my @map_field=("Map_ID");
    my @map_table=("Map");
    my @map_condition=("Map_Version=$map_version");
    my $select_map=SQLdb::GetRecord(\@map_field, \@map_table, \@map_condition);
    my $map_record=SQLdb::ExecuteUnique($dbh, $select_map);
    my @map_arr=@$map_record;

    return $map_arr[0]; ###will return a null value if content 0

}#end sub

#-------------------------------------
sub parseMapXMLStructure{

   my $map_record=shift;
   my @map_record=@$map_record;
   my $maptig_record=shift;
   my $clone_info=shift;
   my $enzyme_info=shift;
   my $xml_record=shift;
   my %xml_record=%$xml_record;
   my $database=shift;
   my $dbh=shift;

   my $map_insert_ct=0;
   my $maptig_insert_ct=0;
   my $layout_insert_ct=0;

   my $map_version=$xml_record{'map_version'};
   my $map_date=$xml_record{'map_date'};
   my $map_comment=$xml_record{'map_comment'};

   my $map_id=checkMapExistence($map_version, $dbh);#call sub to check existence of tig and supertig id

   if (! defined $map_id){####add data for that map version

      my $uniquetable="Map";
      my @map_field=("Map_ID", "Map_Date", "Map_Version", "Map_Comment");
      my @map_value=("", "$map_date", "$map_version", "$map_comment");
      SQLdb::InsertRecord($dbh, $uniquetable, \@map_field, \@map_value);
      $map_id=++$map_record[0];#increment the map record
      $map_insert_ct++;
   }else{####Data already exists for that map version
      die "The Fingerprint Map XML file you specified contains a version of the map (v $map_version) that already exists in the $database database.\n";
   }

   my $ref_tig=$xml_record{'contig'};
   my @arr_of_hash=@$ref_tig;###first layer is map contig

   foreach my $record_reference (@arr_of_hash){

      my $maptig_id=checkMapTigExistence($record_reference, $dbh, $map_id);#call sub to check existence of map contig
      ($maptig_record, $maptig_insert_ct, $layout_insert_ct)=insertMapTigData($maptig_id, $maptig_record, $clone_info, $enzyme_info, $record_reference, $dbh, $map_id, $maptig_insert_ct, $layout_insert_ct);
   }

   return $map_insert_ct, $maptig_insert_ct, $layout_insert_ct;

}#end sub

#-------------------------------------

sub checkMapTigExistence{

        my $record=shift;
        my %record=%$record;
        my $dbh=shift;
        my $map_id=shift;

        #CHECK EXISTENCE OF MAP CONTIG
        my @contig_field=("Map_Contig_ID");
        my @contig_table=("Map_Contig");
        my @contig_condition=("Map_Contig_Number=$record{'contig_number'}", "FK_Map__ID=$map_id");
        my $select_contig=SQLdb::GetRecord(\@contig_field, \@contig_table, \@contig_condition);
        my $tmp_tig=SQLdb::ExecuteUnique($dbh, $select_contig);
        my @tig_existence=@$tmp_tig;
        my $maptig_id=$tig_existence[0];

        return $maptig_id;

}#end sub

#-------------------------------------
sub insertMapTigData{

    my $tig_id=shift;
    my $maptig_record=shift;
    my @maptig_record=@$maptig_record;
    my $clone_info=shift;
    my %clone_info=%$clone_info;
    my $enzyme_info=shift;
    my %enzyme_info=%$enzyme_info;
    my $record=shift;
    my %record=%$record;
    my $dbh=shift;
    my $map_id=shift;
    my $maptig_insert_ct=shift;
    my $layout_insert_ct=shift;

    ####MAP_CONTIG TABLE

    if (! defined $tig_id){#INSERT
       my $uniquetable="Map_Contig";
       my @contig_field=("Map_Contig_ID", "FK_Map__ID", "Map_Contig_Number", "Map_Contig_Size");
       my @contig_value=("", "$map_id", "$record{'contig_number'}", "$record{'contig_size'}");
       SQLdb::InsertRecord($dbh, $uniquetable, \@contig_field, \@contig_value);

       $tig_id=++$maptig_record[0];  # $tig_id=++$maptig_record[0];
       $maptig_insert_ct++;
    }

    ####MAP_LAYOUT TABLE (now this is a bit tricky -- please refer to the data structure)

    my $clone_arr_reference=$record{'clone_name'};
    my @clone_arr_of_hash=@$clone_arr_reference;

    foreach my $clone_reference (@clone_arr_of_hash){#all read of a given contig record

            my %clone=%$clone_reference;
            my $clone_name=$clone{'content'};
            my $clone_altname=$clone{'altname'};
            my $clone_buried=$clone{'buried'};
            my $clone_order=$clone{'order'};
            my $clone_overlap=$clone{'overlap'};
            my $clone_digest=$clone{'digest'};
            my $clone_enzyme=uc($clone{'enzyme'}); ###converts all to upercase

            #CHECK EXISTENCE OF READ IN ASSEMBLED TABLE
            my @layout_field=("Map_Layout_ID");
            my @layout_table=("Map_Layout");
            my @layout_condition=("FK_Map_Contig__ID=$tig_id", "FK_Clone__ID=$clone_info{$clone_name}", "FK_Enzyme__ID=$enzyme_info{$clone_enzyme}");#condition is if foreign keys exist
            my $select_layout=SQLdb::GetRecord(\@layout_field, \@layout_table, \@layout_condition);
            my $tmp_layout=SQLdb::ExecuteUnique($dbh, $select_layout);
            my @layout_existence=@$tmp_layout;
            my $layout_id=$layout_existence[0];#returns unique value, 1st element of array

            if (! defined $layout_id){#INSERT DATA ONLY IF DOESN'T EXIST
               my $uniquetable="Map_Layout";
               my @layout_field=("Map_Layout_ID", "FK_Map_Contig__ID", "FK_Clone__ID", "FK_Enzyme__ID", "Map_Layout_Order", "Map_Layout_Buried", "Map_Layout_Overlap", "Map_Layout_AltCloneName", "Map_Layout_Experimental_Digest");
               my @layout_value=("","$tig_id","$clone_info{$clone_name}", "$enzyme_info{$clone_enzyme}", "$clone_order", "$clone_buried", "$clone_overlap", "$clone_altname", "$clone_digest");
               SQLdb::InsertRecord($dbh, $uniquetable, \@layout_field, \@layout_value);
               $layout_insert_ct++;
            }
        }
        return \@maptig_record, $maptig_insert_ct, $layout_insert_ct;#keep update of clone # for new insertions

}#end sub

###################################################################### assembly.xml parser #############################################################

#-------------------------------------

sub assemblyData{

   my $user_info=shift;
   my %user_info=%$user_info;
   my $xml_dir=shift;
   my $assembly_id=shift;
   my $xml_assembly=shift;
   my $date_condition=shift;

   my $xmlfile="$xml_dir/$xml_assembly";#the xml files will always be stored directly under assmbly id dir
   my ($supertig_record, $tig_record, $read_info);
   my ($supertig_insert_ct, $contig_insert_ct, $assembled_insert_ct);

   ####Initialize Timer
   my $start_date=&getDate();

   ####Get XML data into hashref
   print "Parsing Assembly XML...\n";
   my $xml_record=XMLin($xmlfile, forcearray=>['read_name'], suppressempty=>1);#read_name will always be in array

   #print Dumper($xml_record); 
   #&testStructure($xml_record); 
 
   ####Initialize database connection
   print "Connecting to $user_info{'database'} database...\n";
   my $dbh=SQLdb::Initialize($user_info{'password'}, $user_info{'username'}, $user_info{'database'}, $user_info{'host_name'});

   ####Get Contig, Supertig and Trace initial values
   print "Getting Initial db values...\n";
   ($supertig_record, $tig_record, $read_info)=&getInitialAssemblyValue($dbh, $assembly_id, $date_condition);

   ####Fill database
   print "Populating $user_info{'database'} database...\n";
   ($supertig_insert_ct, $contig_insert_ct, $assembled_insert_ct)=&parseAssemblyXMLStructure($supertig_record, $tig_record, $read_info, $xml_record, $dbh, $assembly_id);

   ####End Time
   my $end_date=&getDate();

   ####Time Elapsed
   my $time_elapsed=&DateCalc($start_date,$end_date);

   print "*started at    :$start_date\n";
   print "*Finished at   :$end_date\n";
   print "*Time Elapsed  :$time_elapsed\n";

   print "Supercontig inserted:$supertig_insert_ct\nContig inserted:$contig_insert_ct\nAssembled inserted:$assembled_insert_ct\n";

   $dbh->disconnect();

}

#-------------------------------------

sub testStructure{

   my $xml_record=shift;
   my %xml=%$xml_record;

   foreach my $type(sort keys %xml){
      print "$type\n";
   }
   
   my $ref_tig=$xml{'contig'};
   my @arr_of_hash=@$ref_tig;
  
   foreach my $ref(@arr_of_hash){
       my %record=%$ref;
       foreach my $rec(sort keys %record){
          print "$rec\t$record{$rec}\n";
       }
   }

}

#-------------------------------------

sub getInitialAssemblyValue{

   my $dbh=shift;
   my $assembly_id=shift;
   my $date_condition=shift;

   my @supertig_field=("count(Supercontig_ID)");
   my @supertig_table=("Supercontig");
   my @supertig_condition=();
   my $select_supertig=SQLdb::GetRecord(\@supertig_field, \@supertig_table, \@supertig_condition);
   my $supertig_record=SQLdb::ExecuteUnique($dbh, $select_supertig);

   my @tig_field=("count(Contig_ID)");
   my @tig_table=("Contig");
   my @tig_condition=();
   my $select_tig=SQLdb::GetRecord(\@tig_field, \@tig_table, \@tig_condition);
   my $tig_record=SQLdb::ExecuteUnique($dbh, $select_tig);

   #GET READ ID (make use of read date time sequenced --->cut on process time)
   my @read_field=("Trace_Name", "Trace_ID");
   my @read_table=("Trace");
   my @read_condition=("$date_condition");
   my $select_read=SQLdb::GetRecord(\@read_field, \@read_table, \@read_condition);
   my $multiple_read=SQLdb::ExecuteMany($dbh, $select_read);
   my @read_record=@$multiple_read;
   my %read_info;

   #print "$select_read\n";
   foreach my $read_rec(@read_record){
      my @read_data=@$read_rec;
      $read_info{$read_data[0]}=$read_data[1];
   #   print "$read_data[0] $read_data[1]\n";
   }

   return $supertig_record, $tig_record, \%read_info;

}

#-------------------------------------

sub getDate{

   my $date=`date`;
   chomp($date);
   return $date;

}

#-------------------------------------
sub parseAssemblyXMLStructure{

   my $supertig_record=shift;
   my $tig_record=shift;
   my $read_info=shift;
   my $xml_record=shift;
   my %xml_record=%$xml_record;
   my $dbh=shift;
   my $assembly_id=shift;
  
   my ($supertig_insert_ct, $contig_insert_ct, $assembled_insert_ct, $pass)=(0,0,0,0);

   my $ref_tig=$xml_record{'contig'};

   eval{
      my @is_array=@$ref_tig;
   };
   if($@){$pass=0;}else{$pass=1;} 


   if(! $pass){###bypass xml structure if single record
   ###one contig
      my $record_reference=$ref_tig;
      my ($supertig_id, $tig_id)=checkTigExistence($record_reference, $dbh, $assembly_id);#call sub to check existence of tig and supertig id
      ($supertig_record, $tig_record, $supertig_insert_ct, $contig_insert_ct, $assembled_insert_ct)=insertTigData($supertig_id, $tig_id, $supertig_record, $tig_record, $read_info, $record_reference, $dbh, $assembly_id, $supertig_insert_ct, $contig_insert_ct, $assembled_insert_ct);
   }else{
      ###many contigs
      my @arr_of_hash=@$ref_tig;
  
      foreach my $record_reference (@arr_of_hash){
         my ($supertig_id, $tig_id)=checkTigExistence($record_reference, $dbh, $assembly_id);#call sub to check existence of tig and supertig id
         ($supertig_record, $tig_record, $supertig_insert_ct, $contig_insert_ct, $assembled_insert_ct)=insertTigData($supertig_id, $tig_id, $supertig_record, $tig_record, $read_info, $record_reference, $dbh, $assembly_id, $supertig_insert_ct, $contig_insert_ct, $assembled_insert_ct); 
      }
   }

  return $supertig_insert_ct, $contig_insert_ct, $assembled_insert_ct;

}#end sub



#-------------------------------------

sub checkTigExistence{

        my $record=shift;
        my %record=%$record;
        my $dbh=shift;
        my $assembly_id=shift;

        #CHECK EXISTENCE OF SUPERCONTIG invoke SQLdb.pm module see module definition
        my $supertig_id='';

        if (defined $record{'supertig_number'}){#this is crucial since some contigs are not found in any supercontigs
           my @supertig_field=("Supercontig_ID");
           my @supertig_table=("Supercontig");
           my @supertig_condition=("Supercontig_Number=$record{'supertig_number'}", "FK_Assembly__ID=$assembly_id"); #assembly id and supercontig number are unique identifiers
           my $select_supertig=SQLdb::GetRecord(\@supertig_field, \@supertig_table, \@supertig_condition);

           my $tmp_supertig=SQLdb::ExecuteUnique($dbh, $select_supertig);       
           my @supertig_existence=@$tmp_supertig;
           $supertig_id=$supertig_existence[0];
        }
        
        #CHECK EXISTENCE OF CONTIG
        my @contig_field=("Contig_ID");
        my @contig_table=("Contig");
        my @contig_condition=("Contig_Number=$record{'contig_number'}", "FK_Assembly__ID=$assembly_id");
        my $select_contig=SQLdb::GetRecord(\@contig_field, \@contig_table, \@contig_condition);
        my $tmp_tig=SQLdb::ExecuteUnique($dbh, $select_contig);
        my @tig_existence=@$tmp_tig;
        my $tig_id=$tig_existence[0];        
       
        return $supertig_id, $tig_id;

}#end sub

#-------------------------------------

sub insertTigData{

        my $supertig_id=shift;
        my $tig_id=shift;
        my $supertig_record=shift;
        my @supertig_record=@$supertig_record;         #Essential to link Contig to Supercontig
        my $tig_record=shift;
        my @tig_record=@$tig_record;                   #Essential to link Assembled to Contig
        my $read_info=shift;
        my %read_info=%$read_info;
        my $record=shift;
        my %record=%$record;
        my $dbh=shift;
        my $assembly_id=shift;
        my $supertig_insert_ct=shift;
        my $contig_insert_ct=shift;
        my $assembled_insert_ct=shift;
        
        #SUPERCONTIG TABLE
        if (! defined $supertig_id){#SUPERCONTIG DOESN'T EXIST IN DATABASE, INSERT DATA FOR THAT SUPERCONTIG
           my $uniquetable="Supercontig";
           my @supertig_field=("Supercontig_ID", "FK_Assembly__ID", "Supercontig_Number", "Supercontig_Size");
           my @supertig_value=("", "$assembly_id", "$record{'supertig_number'}", "$record{'supertig_size'}");
           SQLdb::InsertRecord($dbh, $uniquetable, \@supertig_field, \@supertig_value);
           $supertig_id=++$supertig_record[0];#increment the clone record (essential to link accurate Clone_ID to Trace when Inserting new record)
           $supertig_insert_ct++;
        }
        
        #CONTIG TABLE
        #CONVERT QUALITY & SEQUENCE INTO BINARY AND INSERT IT
        #my $sequence_blob=pack("a*", $record{'contig_sequence'});
        #my @quality=split(' ',$record{'contig_quality'});
        #my $quality_blob=pack("C*", @quality);
        #my @contig_binary=($sequence_blob, $quality_blob);
        #my @contig_binary=("$quality_blob");  

        if (! defined $tig_id){#INSERT
           my $uniquetable="Contig";
           my @contig_field=("Contig_ID", "FK_Assembly__ID", "FK_Supercontig__ID", "Contig_Number", "Contig_Error", "Contig_Size", "Contig_Order", "Contig_PrecedingGap_Size", "Contig_CompA", "Contig_CompT", "Contig_CompG", "Contig_CompC", "Contig_Phred20", "Contig_Phred30","Contig_Phred40");

           #my @contig_field=("Contig_ID", "FK_Assembly__ID", "FK_Supercontig__ID", "Contig_Number", "Contig_Sequence", "Contig_Quality", "Contig_Error", "Contig_Size", "Contig_Order", "Contig_PrecedingGap_Size", "Contig_CompA", "Contig_CompT", "Contig_CompG", "Contig_CompC", "Contig_Phred20", "Contig_Phred30","Contig_Phred40");

           #my @contig_value=("", "$assembly_id", "$supertig_id", "$record{'contig_number'}", "$record{'contig_sequence'}", "?", "$record{'contig_error'}", "$record{'contig_size'}", "$record{'contig_order'}", "$record{'contig_preceding_gap'}", "$record{'contig_compa'}","$record{'contig_compt'}","$record{'contig_compg'}","$record{'contig_compc'}", "$record{'contig_phred20'}", "$record{'contig_phred30'}", "$record{'contig_phred40'}");

           my @contig_value=("", "$assembly_id", "$supertig_id", "$record{'contig_number'}", "$record{'contig_error'}", "$record{'contig_size'}", "$record{'contig_order'}", "$record{'contig_preceding_gap'}", "$record{'contig_compa'}","$record{'contig_compt'}","$record{'contig_compg'}","$record{'contig_compc'}", "$record{'contig_phred20'}", "$record{'contig_phred30'}", "$record{'contig_phred40'}");

           #SQLdb::InsertRecord($dbh, $uniquetable, \@contig_field, \@contig_value, \@contig_binary);
           SQLdb::InsertRecord($dbh, $uniquetable, \@contig_field, \@contig_value);

           $tig_id=++$tig_record[0];  # $tig_id=++$tig_record[0];
           $contig_insert_ct++;
        }

        #ASSEMBLED TABLE (now this is a bit tricky -- please refer to the data structure)
        my $read_arr_reference=$record{'read_name'};
        my @read_arr_of_hash=@$read_arr_reference;
        
        foreach my $read_reference (@read_arr_of_hash){#all read of a given contig record

            my %read=%$read_reference;
            my $read_trace_name=$read{'content'};
            my $read_start=$read{'start'};
            my $read_stop=$read{'stop'};
            my $read_direction=$read{'direction'};

            #CHECK EXISTENCE OF READ IN ASSEMBLED TABLE
            my @assembl_field=("Assembled_ID");
            my @assembl_table=("Assembled");
            my @assembl_condition=("FK_Contig__ID=$tig_id", "FK_Trace__ID=$read_info{$read_trace_name}");#condition is if foreign keys exist
            my $select_assembl=SQLdb::GetRecord(\@assembl_field, \@assembl_table, \@assembl_condition);
            my $tmp_assembl=SQLdb::ExecuteUnique($dbh, $select_assembl);
            my @assembl_existence=@$tmp_assembl;
            my $assembled_id=$assembl_existence[0];#returns unique value, 1st element of array

            #print "$assembled_id -- $select_assembl\n";
 
            if (! defined $assembled_id){#INSERT DATA ONLY IF DOESN'T EXIST
               my $uniquetable="Assembled";
               my @assembled_field=("Assembled_ID", "FK_Contig__ID", "FK_Trace__ID", "Assembled_Read_Start", "Assembled_Read_Stop", "Assembled_Direction");
               my @assembled_value=("","$tig_id","$read_info{$read_trace_name}", "$read_start", "$read_stop", "$read_direction");
               SQLdb::InsertRecord($dbh, $uniquetable, \@assembled_field, \@assembled_value);  
               $assembled_insert_ct++;
               
            }
        }
        return \@supertig_record, \@tig_record, $supertig_insert_ct, $contig_insert_ct, $assembled_insert_ct;#keep update of clone # for new insertions

}#end sub

##############################################Summary.xml Parser #############################################################

#-------------------------------------
#MUST INCLUDE SOME ERROR TRAPPING IN HERE IN CASE THE XML FILE ISN'T IN RIGHT FORMAT
sub summaryData{

   my $user_info=shift;
   my %user_info=%$user_info;
   my $xml_dir=shift;
   my $assembly_id=shift;
   my $xml_assembly=shift;

   my $xmlfile="$xml_dir/$xml_assembly";#the xml files will always be stored directly under assmbly id dir

   ####Initialize Timer
   my $start_date=&getDate();

   ####Get XML data into hashref
   print "Parsing Summary XML...\n";
   my $xml_record=XMLin($xmlfile, forcearray=>0, suppressempty=>1);

   #print Dumper($xml_record);

   ####Initialize database connection
   print "Connecting to $user_info{'database'} database...\n";
   my $dbh=SQLdb::Initialize($user_info{'password'}, $user_info{'username'}, $user_info{'database'}, $user_info{'host_name'});

   ####Fill assembly database
   print "Populating $user_info{'database'} database...\n";
   &parseSummaryXMLStructure($xml_record, $dbh, $assembly_id);

   ####End Time
   my $end_date=&getDate();

   ####Time Elapsed
   my $time_elapsed=&DateCalc($start_date,$end_date);
   
   print "*started at    :$start_date\n";
   print "*Finished at   :$end_date\n";
   print "*Time Elapsed  :$time_elapsed\n";

      
   $dbh->disconnect();

}

#-------------------------------------

sub parseSummaryXMLStructure{

   my $xml_record=shift;
   my %xml_record=%$xml_record;
   my $dbh=shift;
   my $assembly_id=shift;

   my $record_reference=$xml_record{'assembly_summary'};

   my ($summary_id, $confirmed_assembly_id)=&checkSummaryExistence($record_reference, $dbh, $assembly_id);#call sub to check existence of tig and supertig id
   &insertSummaryData($record_reference, $summary_id, $confirmed_assembly_id, $dbh);


}

#-------------------------------------
sub checkSummaryExistence{

        my $record=shift;
        my %record=%$record;
        my $dbh=shift;
        my $assembly_id=shift;

        #CHECK EXISTENCE OF SUMMARY #invoke SQLdb.pm module see module definition
        my @summary_field=("Summary_ID");
        my @summary_table=("Summary");
        my @summary_condition=("FK_Assembly__ID=$assembly_id"); #assembly id is unique identifiers
        my $select_summary=SQLdb::GetRecord(\@summary_field, \@summary_table, \@summary_condition);
        my $tmp_summary=SQLdb::ExecuteUnique($dbh, $select_summary);       
        my @summary_existence=@$tmp_summary;
        my $summary_id=$summary_existence[0];

        #CHECK EXISTENCE OF ASSEMBLY
        my @assembly_field=("Assembly_ID");
        my @assembly_table=("Assembly");
        my @assembly_condition=("Assembly_ID=$assembly_id"); #assembly id is unique identifiers
        my $select_assembly=SQLdb::GetRecord(\@assembly_field, \@assembly_table, \@assembly_condition);
        my $tmp_assembly=SQLdb::ExecuteUnique($dbh, $select_assembly);
        my @assembly_existence=@$tmp_assembly;
        my $confirmed_assembly_id=$assembly_existence[0];

        if (! defined $assembly_id){
           $dbh->disconnect();
           die "Can't use an undefined foreign key for @assembly_field\n";
        }

        return $summary_id, $confirmed_assembly_id;

}#end sub

#-------------------------------------
sub insertSummaryData{

      my $record=shift;
      my %record=%$record;
      my $summary_id=shift;
      my $confirmed_assembly_id=shift;
      my $dbh=shift;

      #SUMMARY TABLE
      if (! defined $summary_id){#This is unlikely since an entry must exist at this point
         #SUMMARY TABLE
         my $uniquetable="Summary";
         my @summary_field=("Summary_ID", "FK_Assembly__ID", "Summary_Contig_Number", "Summary_Supertig_Number", "Summary_Supertig_Size", "Summary_Singleton_Read", "Summary_Bad_Quality_Read", "Summary_Masked_Read", "Summary_Error", "Summary_Largest_Tig", "Summary_Smallest_Tig", "Summary_Mean_Tig", "Summary_Mean_Super", "Summary_Total_Bases", "Summary_Consensus", "Summary_Phred30", "Summary_Phred40", "Summary_Above1K_Tigsize", "Summary_N50", "Summary_Percent_A", "Summary_Percent_T", "Summary_Percent_C", "Summary_Percent_G", "Summary_Assembled_Read", "Summary_Singleton_Contig", "Summary_2Read_Contig", "Summary_3Read_Contig", "Summary_4Read_Contig", "Summary_5Read_Contig", "Summary_6Read_Contig", "Summary_10Read_Contig", "Summary_20Read_Contig", "Summary_Genome_Coverage", "Summary_Contig_Coverage", "Summary_Poisson_Genome_Size", "Summary_Poisson_Genome_Coverage", "Summary_Poisson_Contig", "Summary_Genome_Coverage2", "Summary_Poisson_Genome_Coverage2", "Summary_Analyzed_Read");
         my @summary_value=("", "$confirmed_assembly_id", "$record{'contig_number'}", "$record{'supertig_number'}", "$record{'supertig_size'}", "$record{'singleton_read'}", "$record{'badquality_read'}", "$record{'masked_read'}", "$record{'average_error'}", "$record{'biggest_contig'}","$record{'smallest_contig'}", "$record{'mean_tigsize'}","$record{'mean_supersize'}", "$record{'total_bp'}","$record{'consensus'}","$record{'phred30'}", "$record{'phred40'}","$record{'above1kcontig_size'}", "$record{'N50_length'}", "$record{'percent_a'}","$record{'percent_t'}", "$record{'percent_g'}","$record{'percent_c'}", "$record{'read_assembled'}", "$record{'contig_singleton'}", "$record{'two_read'}", "$record{'three_read'}", "$record{'four_read'}", "$record{'five_read'}", "$record{'six_read'}", "$record{'ten_read'}", "$record{'twenty_read'}", "$record{'genome_coverage'}", "$record{'contig_coverage'}", "$record{'poisson_genome'}", "$record{'poisson_coverage'}", "$record{'poisson_contig'}", "$record{'genome_coverage2'}", "$record{'poisson_coverage2'}", "$record{'read_analyzed'}");
         SQLdb::InsertRecord($dbh, $uniquetable, \@summary_field, \@summary_value);
     }   
     else{#update
         print "Summary table updated.\n";
         my $uniquetable="Summary";
         my @summary_field=("Summary_ID", "Summary_Contig_Number", "Summary_Supertig_Number", "Summary_Supertig_Size", "Summary_Singleton_Read", "Summary_Bad_Quality_Read", "Summary_Masked_Read", "Summary_Error", "Summary_Largest_Tig", "Summary_Smallest_Tig", "Summary_Mean_Tig", "Summary_Mean_Super", "Summary_Total_Bases", "Summary_Consensus", "Summary_Phred30", "Summary_Phred40", "Summary_Above1K_Tigsize", "Summary_N50", "Summary_Percent_A", "Summary_Percent_T", "Summary_Percent_C", "Summary_Percent_G", "Summary_Assembled_Read", "Summary_Singleton_Contig", "Summary_2Read_Contig", "Summary_3Read_Contig", "Summary_4Read_Contig", "Summary_5Read_Contig", "Summary_6Read_Contig", "Summary_10Read_Contig", "Summary_20Read_Contig", "Summary_Genome_Coverage", "Summary_Contig_Coverage", "Summary_Poisson_Genome_Size", "Summary_Poisson_Genome_Coverage", "Summary_Poisson_Contig", "Summary_Genome_Coverage2", "Summary_Poisson_Genome_Coverage2", "Summary_Analyzed_Read");
         my @summary_value=("$summary_id", "$record{'contig_number'}", "$record{'supertig_number'}", "$record{'supertig_size'}", "$record{'singleton_read'}", "$record{'badquality_read'}", "$record{'masked_read'}", "$record{'average_error'}", "$record{'biggest_contig'}","$record{'smallest_contig'}", "$record{'mean_tigsize'}","$record{'mean_supersize'}", "$record{'total_bp'}","$record{'consensus'}", "$record{'phred30'}", "$record{'phred40'}","$record{'above1kcontig_size'}", "$record{'N50_length'}", "$record{'percent_a'}","$record{'percent_t'}", "$record{'percent_c'}","$record{'percent_g'}", "$record{'read_assembled'}", "$record{'contig_singleton'}", "$record{'two_read'}", "$record{'three_read'}", "$record{'four_read'}", "$record{'five_read'}", "$record{'six_read'}", "$record{'ten_read'}", "$record{'twenty_read'}", "$record{'genome_coverage'}", "$record{'contig_coverage'}", "$record{'poisson_genome'}", "$record{'poisson_coverage'}", "$record{'poisson_contig'}", "$record{'genome_coverage2'}", "$record{'poisson_coverage2'}", "$record{'read_analyzed'}");
         my @summary_condition=("FK_Assembly__ID=$confirmed_assembly_id");#don't want to use the table id here, because if name change, the whole process of analysis changes.
         SQLdb::UpdateRecord($dbh, $uniquetable, \@summary_field, \@summary_value, \@summary_condition);
     }

}#end sub

################################################ Parser for XML trace Archive Info #####################################################
#This Code won't work if xml file is too big (because XML::simple populate all in an hash).  Tha's why xmltrace_parser.pl was developed 
########################################################################################################################################


#-------------------------------------
sub readData{

   my $user_info=shift;
   my %user_info=%$user_info;
   my $xml_dir=shift;
   my $assembly_id=shift;
   my $xml_assembly=shift;

   my $xmlfile="$xml_dir/$xml_assembly";#the xml files will always be stored directly under assmbly id dir
   my ($clone_insert_ct, $clone_update_ct,  $read_insert_ct, $read_update_ct);

   ####Initialize Timer
   my $start_date=&getDate();

   ####Get XML data into hashref
   print "Parsing XML...\n";
   my $xml_record=XMLin($xmlfile, forcearray=>0, suppressempty=>1);

   #print Dumper($xml_record);

   ####Initialize database connection
   print "Connecting to $user_info{'database'} database...\n";
   my $dbh=SQLdb::Initialize($user_info{'password'}, $user_info{'username'}, $user_info{'database'}, $user_info{'host_name'});

   ####Get Contig, Supertig and Trace initial values
   print "Getting Initial db values...\n";
   my $clone_record=&getInitialCloneValue($dbh);

   ####Fill assembly database
   print "Populating $user_info{'database'} database...\n";
   ($clone_insert_ct, $clone_update_ct,  $read_insert_ct, $read_update_ct)=&parseTraceXMLStructure($xml_record, $clone_record, $dbh);

   ####End Time
   my $end_date=&getDate();

   ####Time Elapsed
   my $time_elapsed=&DateCalc($start_date,$end_date);

   print "*started at    :$start_date\n";
   print "*Finished at   :$end_date\n";
   print "*Time Elapsed  :$time_elapsed\n";

   print "Read inserted:$read_insert_ct\nRead updated:$read_update_ct\nClone inserted:$clone_insert_ct\nClone updated:$clone_update_ct\n";

   $dbh->disconnect();

}

#-------------------------------------
sub getInitialCloneValue{

   my $dbh=shift;

   my @field=("count(Clone_ID)");
   my @table=("Clone");
   my @condition=();
   my $select_clone=SQLdb::GetRecord(\@field, \@table, \@condition);
   my $clone_record=SQLdb::ExecuteUnique($dbh, $select_clone);

   return $clone_record;

}

#-------------------------------------
sub parseTraceXMLStructure{

   my $xml_record=shift;
   my %xml_record=%$xml_record;
   my $clone_record=shift;
   my $dbh=shift;

   my ($clone_insert_ct, $clone_update_ct, $read_insert_ct, $read_update_ct)=(0,0,0,0);

   my $ref_trace=$xml_record{'assembly_summary'};
   my @arr_of_hash=@$ref_trace;#this is a problem with XML::Simple???!

   foreach my $record_reference (@arr_of_hash){
      my ($clone_ref, $read_ref)=&checkTraceExistence($record_reference, $dbh);#call sub to check existence of clone and read
      ($clone_record, $clone_insert_ct, $clone_update_ct,  $read_insert_ct, $read_update_ct)=&insertTraceData($record_reference, $clone_record, $clone_ref, $read_ref, $clone_insert_ct, $clone_update_ct, $read_insert_ct, $read_update_ct, $dbh);
   }

   return $clone_insert_ct, $clone_update_ct,  $read_insert_ct, $read_update_ct;

}

#-------------------------------------
sub checkTraceExistence{

    my $record=shift;
    my %record=%$record;
    my $dbh=shift;

    #CHECK EXISTENCE OF CLONE #invoke SQLassembly.pm module see module definition
    my @field=("Clone_ID", "Clone_Name");
    my @table=("Clone");
    my @condition=("Clone_Name like '%$record{'clone_id'}%'");
    my $select_clone=SQLdb::GetRecord(\@field, \@table, \@condition);
    my $tmp_clone=SQLdb::ExecuteUnique($dbh, $select_clone);

    #CHECK EXISTENCE OF READ
    @field=("Trace_ID");
    @table=("Trace");
    @condition=("Trace_Name like '%$record{'trace_name'}%'");
    my $select_read=SQLdb::GetRecord(\@field, \@table, \@condition);
    my $tmp_read=SQLdb::ExecuteUnique($dbh, $select_read);

    return $tmp_clone, $tmp_read;

}


#---------------------------
sub insertTraceData{

     my $record=shift;
     my %record=%$record;
     my $clone_record=shift;
     my @clone_record=@$clone_record;#useful if Clone doesn't exists
     my $tmp_clone=shift;
     my @clone_existence=@$tmp_clone;
     my $tmp_read=shift;
     my @read_existence=@$tmp_read;
     my $clone_insert_ct=shift;
     my $clone_update_ct=shift;
     my $read_insert_ct=shift;
     my $read_update_ct=shift;
     my $dbh=shift;

     my ($clone_id, $clone_name);
     ($clone_id, $clone_name)=@clone_existence;
     my $read_name=$read_existence[0];

     $dbh->{RaiseError}=1;

     if (! defined $clone_id){#CLONE DOESN'T EXIST IN DATABASE, INSERT DATA FOR THAT CLONE
        #CLONE TABLE
        my $uniquetable="Clone";
        my @field=("Clone_ID", "Clone_Name", "Clone_Insert_Size", "Clone_Insert_Stdev", "Clone_Library");
        my @value=("", "$record{'clone_id'}", "$record{'insert_size'}", "$record{'insert_stdev'}", "$record{'library_id'}");
        SQLdb::InsertRecord($dbh, $uniquetable, \@field, \@value);
        $clone_record[0]++;#increment the clone record (essential to link accurate Clone_ID to Trace when Inserting new record)
        $clone_id=$clone_record[0];
        $clone_insert_ct++;
     }#No Clone Updates at this point-->would take too long

     #READ TABLE
     #CONVERT QUALITY INTO BINARY AND INSERT IT
     my @quality=split (' ',$record{'trace_quality'});
     my $quality_blob=pack("C*", @quality);
     my @binary=($quality_blob);

     if (! defined $read_existence[0]){#INSERT

        my $uniquetable="Trace";
        my @field=("Trace_ID", "FK_Clone__ID", "Trace_Name", "Trace_Directory", "Trace_Chemistry", "Trace_Template", "Trace_Well", "Trace_Plate_Number", "Trace_Plate_Quadrant", "Trace_Type", "Trace_sequence", "Trace_quality", "Trace_Length", "Trace_Date_Sequenced");
        my @value=("", "$clone_id", "$record{'trace_name'}", "$record{'trace_directory'}", "$record{'chemistry'}", "$record{'template_id'}", "$record{'well_id'}", "$record{'plate_number'}", "$record{'plate_quadrant'}", "$record{'type'}", "$record{'trace_sequence'}","?","$record{'trace_length'}", "$record{'run_date'}");
        SQLdb::InsertRecord($dbh, $uniquetable, \@field, \@value, \@binary);
        $read_insert_ct++;

     }
     else{#UPDATE

        print "$read_name\t$record{'trace_name'} updated.\n";
        my $uniquetable="Trace";
        my @field=("FK_Clone__ID", "Trace_Name", "Trace_Directory", "Trace_Chemistry", "Trace_Well","Trace_Plate_Number", "Trace_Plate_Quadrant", "Trace_Type", "Trace_Sequence", "Trace_Quality", "Trace_Length","Trace_Date_Sequenced");
        my @value=("$clone_id", "$record{'trace_name'}", "$record{'trace_directory'}", "$record{'chemistry'}", "$record{'well_id'}", "$record{'plate_number'}", "$record{'plate_quadrant'}", "$record{'type'}", "$record{'trace_sequence'}", "?", "$record{'trace_length'}","$record{'run_date'}");
        my @condition=("Trace_Name like '%$record{'trace_name'}%'");#don't want to use the table id here, because if name change, the whole process of analysis changes.
        SQLdb::UpdateRecord($dbh, $uniquetable, \@field, \@value, \@condition, \@binary);
        $read_update_ct++;
     }
     return \@clone_record, $clone_insert_ct, $clone_update_ct,  $read_insert_ct, $read_update_ct;#keep update of clone # for new insertions

}#end sub

1;
